# Copyright: 2007-2011 Paul Obermeier (obermeier@poSoft.de)
# Distributed under BSD license.

package provide tcomexcel 0.4.0

namespace eval ::Excel {

    # Find a string in a cell range. If endRow and/or endCol is negative,
    # all rows/columns are searched. The first cell matching the search
    # string is returned as a 2-element list {row, col}.
    proc FindCell { cellsId str { startRow 1 } { endRow -1 } \
                                { startCol 1 } { endCol -1 } } {
        if { $endRow < 0 } {
            set endRow [::Excel::GetNumRows $cellsId]
        }
        if { $endCol < 0 } {
            set endCol [::Excel::GetNumColumns $cellsId]
        }
        for { set row $startRow } { $row <= $endRow } { incr row } {
            for { set col $startCol } { $col <= $endCol } { incr col } {
                set cellValue [::Excel::GetCellValue $cellsId $row $col]
                if { $cellValue eq $str } {
                    return [list $row $col]
                }
            }
        }
        return [list -1 -1]
    }

    proc SetHeaderLine { cellsId headerList { row 1 } { startCol 1 } } {
        set len [llength $headerList]
        ::Excel::SetRowValues $cellsId $row $startCol $headerList $len
        ::Excel::FormatHeaderLine $cellsId $startCol [expr {$startCol + $len}] $row
    }

    proc FormatHeaderLine { cellsId startCol endCol { row 1 } } {
        set header [::Excel::SelectRangeByIndex $cellsId $row $startCol \
                                                         $row $endCol]
        ::Excel::SetRangeHorizontalAlignment $header $::Excel::xlCenter
        ::Excel::SetRangeVerticalAlignment   $header $::Excel::xlCenter
        ::Excel::SetRangeFontBold $header
    }

    proc CopyWorksheet { fromWorksheetId toWorksheetId } {
        $fromWorksheetId Activate
        set rangeId [::Excel::SelectAll $fromWorksheetId]
        $rangeId Copy

        $toWorksheetId Activate
        $toWorksheetId Paste
    }

    # Insert the matrix values contained in "matrixList" into the worksheet
    # identified by "worksheetId".
    # The insertion of the matrix data starts at cell "startRow,startCol".
    # The matrix data must be stored as a list of lists. Each sub-list contains
    # the values for the row values.
    # Note: This procedure uses the Twapi extension for fast insertion of large 
    #       data. If Twapi is not available, it uses the SetRowValues 
    #       procedure and is significantly slower then.
    proc SetMatrixValues { worksheetId matrixList { startRow 1 } { startCol 1 } } {
        variable useTwapi
        variable mapFloatPoint

        set cellsId [::Excel::GetCellsId $worksheetId]
        if { [::Excel::HavePkg "twapi"] && $useTwapi } {
            set csvFmt [twapi::register_clipboard_format "Csv"]

            twapi::open_clipboard
            twapi::empty_clipboard
            ::Csv::SetSeparatorChar ";"
            twapi::write_clipboard $csvFmt [::Csv::List2Csv $matrixList $mapFloatPoint]
            twapi::close_clipboard

            set cellId [::Excel::SelectRangeByIndex $cellsId \
                        $startRow $startCol $startRow $startCol true]

            # If not waiting at least a little while, 
            # on some computers the Paste method fails.
            after 10
            $worksheetId Paste
        } else {
            set curRow $startRow
            foreach rowList $matrixList {
                ::Excel::SetRowValues $cellsId $curRow $startCol $rowList
                incr curRow
            }
        }
    }

    proc GetMatrixValues { worksheetId row1 col1 row2 col2 } {
        variable useTwapi
        variable mapFloatPoint

        set cellsId [::Excel::GetCellsId $worksheetId]
        if { [::Excel::HavePkg "twapi"] && $useTwapi } {
            set rangeId [::Excel::SelectRangeByIndex $cellsId \
                         $row1 $col1 $row2 $col2 true]
            $rangeId Copy

            set csvFmt [twapi::register_clipboard_format "Csv"]
            while { ! [twapi::clipboard_format_available $csvFmt] } {
                after 10
            }
            twapi::open_clipboard
            set clipboardData [twapi::read_clipboard $csvFmt]
            twapi::close_clipboard

            ::Csv::SetSeparatorChar ";"
            set matrixList [::Csv::Csv2List $clipboardData $mapFloatPoint]
        } else {
            set numVals [expr {$col2-$col1+1}]
            for { set row $row1 } { $row <= $row2 } { incr row } {
                lappend matrixList [::Excel::GetRowValues $cellsId $row $col1 $numVals]
            }
        }
        return $matrixList
    }

    proc TablelistToWorksheet { tableId worksheetId { useHeader true } } {
        set cellsId [::Excel::GetCellsId $worksheetId]
        set startRow 1
        if { $useHeader } {
            # -align -labelalign
            # -width -1 pixel 0 dynamic
            # -title
            set numCols [$tableId columncount]
            for { set col 0 } { $col < $numCols } { incr col } {
                lappend headerList [$tableId columncget $col -title]
            }
            ::Excel::SetHeaderLine $cellsId $headerList
            incr startRow
        }
        set matrixList [$tableId get 0 end]
        ::Excel::SetMatrixValues $worksheetId $matrixList $startRow 1
    }

    proc WorksheetToTablelist { worksheetId tableId { useHeader true } } {
        set cellsId [::Excel::GetCellsId $worksheetId]
        set numRows [::Excel::GetNumUsedRows $worksheetId]
        set numCols [::Excel::GetNumUsedColumns $worksheetId]
        set startRow 1
        if { $useHeader } {
            set headerList [::Excel::GetRowValues $cellsId 1 1 $numCols]
            foreach title $headerList {
                $tableId insertcolumns end 0 $title left
            }
            incr startRow
        } else {
            for { set col 1 } { $col <= $numCols } { incr col } {
                $tableId insertcolumns end 0 "NN" left
            }
        }
        set excelList [::Excel::GetMatrixValues $worksheetId $startRow 1 $numRows $numCols]
        foreach rowList $excelList {
            $tableId insert end $rowList
        }
    }

    # Copy the data contained in Word table "tableId" into Excel worksheet
    # identified by "worksheetId".
    proc WordTableToWorksheet { tableId worksheetId { useHeader true } } {
        set cellsId [::Excel::GetCellsId $worksheetId]
        set startRow 1
        if { $useHeader } {
            set numCols [::Word::GetNumColumns $tableId]
            for { set col 1 } { $col <= $numCols } { incr col } {
                lappend headerList [::Word::GetCellValue $tableId $startRow $col]
            }
            ::Excel::SetHeaderLine $cellsId $headerList
            incr startRow
        }
        set numRows [::Word::GetNumRows $tableId]
        if { $useHeader } {
            incr numRows -1
        }
        set tableList [::Word::GetMatrixValues $tableId \
                      $startRow 1 [expr {$startRow + $numRows-1}] $numCols]
        ::Excel::SetMatrixValues $worksheetId $tableList $startRow 1
    }

    proc WorksheetToWordTable { worksheetId tableId { useHeader true } } {
        set cellsId [::Excel::GetCellsId $worksheetId]
        set numRows [::Excel::GetNumUsedRows $worksheetId]
        set numCols [::Excel::GetNumUsedColumns $worksheetId]
        set startRow 1
        set headerList [::Excel::GetRowValues $cellsId 1 1 $numCols]
        if { [llength $headerList] < $numCols } {
            set numCols [llength $headerList]
        }
        if { $useHeader } {
            ::Word::SetHeaderLine $tableId $headerList
            incr startRow
        }
        set excelList [::Excel::GetMatrixValues $worksheetId $startRow 1 $numRows $numCols]
        ::Word::SetMatrixValues $tableId $excelList $startRow 1
    }

    proc InsertTestData { cellsId timeHeader timeList valsHeaderList valsList } {
        # Insert first row with header names.
        # First column contains the time header, next columns of row 1
        # contain the header of the data columns.
        SetCellValue $cellsId 1 1 $timeHeader

        set col 2
        foreach head $valsHeaderList {
            SetCellValue $cellsId 1 $col $head
            incr col
        }
        set lastDataCol [expr $col - 1]

        # Format the header lines.
        FormatHeaderLine $cellsId 1 [expr $col-1]

        # Now insert the data row by row.
        set row 2
        foreach t $timeList vals $valsList {
            set col 1
            SetCellValue $cellsId $row $col $t
            incr col
            foreach val $vals {
                SetCellValue $cellsId $row $col $val
                incr col
            }
            incr row
        }
        set lastDataRow [expr $row - 1]

        incr row
        set r $row
        set funcList [list \
            [::Excel::GetLangFuncName "MIN"] \
            [::Excel::GetLangFuncName "MAX"] \
            [::Excel::GetLangFuncName "AVERAGE"] \
            [::Excel::GetLangFuncName "STDEV"]]
        foreach labelStr $funcList {
            SetCellValue $cellsId $r 1 $labelStr
            SetRangeFontBold [::Excel::GetCellIdByIndex $cellsId $r 1] true
            incr r
        }

        for { set c 2 } { $c <= $lastDataCol } { incr c } {
            set dataRange [::Excel::GetCellRange 2 $c $lastDataRow $c]
            set r $row
            foreach func $funcList {
                set cell [SelectRangeByIndex $cellsId $r $c $r $c]
                set formula [format "=%s(%s)" $func $dataRange]
                $cell Formula $formula
                incr r
            }
        }
    }
}
